package FI2.Stream;
/**{p}
 * -- FI2.Stream --
 * Universita' di Padova -
 * Diploma Universitario in Ingegneria Informatica
 * Corso di Fondamenti di Informatica 2
 * Flussi
 */

/**{f}
 * FileMem.java
 * memoria virtuale su file ad accesso casuale
 *
 * @version 1.00 98/10/02
 * @version 1.01 99/10/27 -
 * errore in getValue e setValue
 * @version 2.00 00/09/19 -
 * InvalidArgumentException -> IllegalArgumentException
 * @author M.Moro DEI UPD
 */

import java.io.*;
import FI2.Util.FI2Sys;

/**{c}
 * realizza una specie di memoria virtuale su file in
 * cui e` possibile allocare locazioni da 8 (fByte),
 * 16 (fWord) e 32 bit (fDWord) visti come registri
 * unsigned -
 * la tabella di mappa, che fornisce le posizioni libere
 * o occupate, occupa a sua volta la prima parte del
 * file di supporto -
 * ogni suo byte rappresenta lo stato di 8 locazioni da
 * un byte (1 bit/byte)
 */
public class FileMem
{
    private static final int MAPEXP = 3;
    private static final int MAPELEMSIZE = 8;
    private static final int MAPMASK = 0x7;
      // 3 bit per rappresentare una posizione
      // in una locazione della mappa (che appunto
      // ha locazioni byte 2^3=8)
    static final int BYTESIZE = 1;
    private static final int BYTEMASK = 1;
    static final int WORDSIZE = 2;
    private static final int WORDMASK = 3;
    static final int DWORDSIZE = 4;
    private static final int DWORDMASK = 0xf;

    private RandomAccessFile file;
      // file di supporto
    byte map[];  // mappa per allocazione
    int memSize;  // estensione memoria in byte
    int mapSize;  // estensione mappa in byte

    /**[c]
     * costruttore base -
     * apre il file di supporto e vi legge, se
     * presente, la mappa di allocazione
     * @param filename  nome file di supporto
     * @param sizeExp  esponente di base 2 della dimensione
     *                 in byte della memoria
     * @exception IllegalArgumentException
     * se sizeExp negativo o maggiore di 31
     */
    public FileMem(String filename, int sizeExp)
      throws IOException
    {
        if (sizeExp >= 32 || sizeExp <= 0)
          // size illegale
            throw new IllegalArgumentException(
              "FileMap esponente di size " + sizeExp + 
              " illegale");
        file = new RandomAccessFile(filename, "rw");
        memSize = 1 << sizeExp;
        mapSize = memSize >> MAPEXP;
        map = new byte[mapSize];
        if (file.length() >= mapSize)
        {
            // la mappa c'e` gia`: carica
            for (int i=0; i<mapSize;
              map[i++] = file.readByte());
        }
        else
            for (int i=0; i<mapSize; map[i++] = 0);
              // init map, tutte le locazioni sono libere
    } //[c] FileMem

    /**[m]
     * chiude il file di supporto
     */
    protected void finalize()
      throws IOException
    { file.close(); }

    /**[m]
     * ricerca la prima posizione libera per una
     * variabile di dimensione fornita guardando la
     * locazione di mappa di indice fornito e tenendo
     * conto dell'allineamento (su frontiera 2 per word,
     * su frontiera 4 per dword) -
     * a causa dell'allineamento non puo` accadere che
     * una variabile word o dword abbia uno stato di
     * allocazione memorizzato in piu` locazioni della mappa
     * @param mapPos  indice della locazione di mappa
     * @param size  dimensione in byte della variabile
     * @return indice se trovato 'buco' libero,
     *         -1 altrimenti
     */
    private int getIdx(int mapPos, int size)
    {
        int mask = 0;
        switch (size)
        {
        case BYTESIZE:
          // fByte
            mask = BYTEMASK;
            break;
        case WORDSIZE:
          // fWord
            mask = WORDMASK;
            break;
        case DWORDSIZE:
          // fDWord
            mask = DWORDMASK;
        } //switch

        // l'incremento dell'indice tiene conto
        // dell'allineamento
        for (int idx=0; idx <= MAPELEMSIZE-size; idx+=size)
            if ((map[mapPos] & (mask << idx)) == 0)
              // trovato spazio libero sufficiente
              return idx;
        return -1;
    } //[m] getIdx

    /**[m]
     * alloca una nuova variabile, cercando una
     * posizione libera in base all'allineamento
     * @param size  dimensione in byte della variabile
     * @return indirizzo virtuale della variabile
     * @exception IOException
     * oltre alle altre cause, se non trova locazioni
     * libere
     */
    int newLoc(int size)
      throws IOException
    {
        //si assume size corretto
        int mapPos, mapIdx;

        for (int i=1; i<=3; i++)
        {
            // tenta 3 volte di trovare spazio
            for (mapPos=0; mapPos<mapSize; mapPos++)
            {
                if ((mapIdx = getIdx(mapPos, size))
                  != -1)
                  // trovato un buco
                {
                    switch (size)
                    {
                    case BYTESIZE:
                      // fByte
                        map[mapPos] |= (BYTEMASK<<mapIdx);
                        break;
                    case WORDSIZE:
                      // fWord
                        map[mapPos] |= (WORDMASK<<mapIdx);
                        break;
                    case DWORDSIZE:
                      // fDWord
                        map[mapPos] |=
                          (DWORDMASK<<mapIdx);
                    } //switch
                    file.seek(mapPos);
                    file.writeByte((int)map[mapPos]);
                    return (mapPos << MAPEXP) + mapIdx;
                } // if
                // non ci sono buchi
                System.gc();
                  // tentativo di forzare garbagecollector
            } //for mapPos
        } //for i
        throw new java.io.IOException("Memoria esaurita");
    } //[m] newLoc

    /**[m]
     * dealloca una variabile, aggiornando la mappa
     * @param adr  indirizzo virtuale della variabile
     * @param size  dimensione in byte della variabile
     */
    void freeLoc(int adr, int size)
      throws IOException
    {
        int mapPos = adr >> MAPEXP;
        int mapIdx = adr & MAPMASK;
        // non controlla se gia` liberato
        switch (size)
        {
        case BYTESIZE:
          // fByte
            map[mapPos] &= ~(BYTEMASK << mapIdx);
            break;
        case WORDSIZE:
          // fWord
            map[mapPos] &= ~(WORDMASK << mapIdx);
            break;
        case DWORDSIZE:
          // fDWord
            map[mapPos] &= ~(DWORDMASK << mapIdx);
        } //switch
        file.seek(mapPos);
        file.writeByte((int)map[mapPos]);
    } //[m] freeLoc

    /**[m]
     * dealloca tutte le variabili, azzerando al mappa
     */
    void freeAll()
      throws IOException
    {
        for (int i=0; i<mapSize; )
        {
            // init map, tutte le locazioni sono libere
            file.seek(i);
            file.writeByte((int)(map[i++] = 0));
        } //for
    }//[m] freeAll

    /**[m]
     * lettura di un valore in variabile
     * @param adr  indirizzo virtuale della variabile
     * @param size  dimensione in byte della variabile
     * @return valore letto
     */
    long getValue(int adr, int size)
      throws IOException
    {
/* v1.01 errore
        file.seek((adr >> MAPEXP) + mapSize);
*/
        file.seek(adr + mapSize);
        switch (size)
        {
        case BYTESIZE:
          // fByte
            return (long)file.readUnsignedByte();
        case WORDSIZE:
          // fWord
            return (long)file.readUnsignedShort();
        case DWORDSIZE:
          // fDWord
            return ((long)file.readInt()) & 0xffffffff;
        } //switch
        return 0;
    } //[m] getValue

    /**[m]
     * scrittura di un valore in variabile
     * @param adr  indirizzo virtuale della variabile
     * @param size  dimensione in byte della variabile
     * @param val  valore
     */
    void setValue(int adr, int size, long val)
      throws IOException
    {
/* v1.01 errore
        file.seek((adr >> MAPEXP) + mapSize);
*/
        file.seek(adr + mapSize);
        switch (size)
        {
        case BYTESIZE:
          // fByte
            file.writeByte((int)val);
            break;
        case WORDSIZE:
          // fWord
            file.writeShort((int)val);
            break;
        case DWORDSIZE:
          // fDWord
            file.writeInt((int)val);
        } //switch
    } //[m] setValue

    /**[m][s]
     * metodo di collaudo
     * @param args[0]  opzionale: nome file di supporto
     */
    public static void main(String[] args)
      throws IOException
    {
        if (args.length == 0)
        {
            System.err.println(
"Memoria virtuale, uso: FileMem <pathname file supporto>");
            return;
        }

        FileMem myfm = new FileMem(args[0], 5);
        int adr1;
        FI2Sys.out.println("1 alloca DWord = " + 
          (adr1=myfm.newLoc(4)));
        int adr2;
        FI2Sys.out.println("2 alloca Word = " + 
          (adr2=myfm.newLoc(2)));
        int adr3;
        FI2Sys.out.println("3 alloca Byte = " + 
          (adr3=myfm.newLoc(1)));

        myfm.setValue(adr1, 4, 1);
        FI2Sys.out.println("1 DWord val (1)= " + 
          myfm.getValue(adr1, 4));
        myfm.setValue(adr2, 2, -1);
        FI2Sys.out.println("2 Word val (-1)= " + 
          myfm.getValue(adr2, 2));
        myfm.setValue(adr3, 1, 250);
        FI2Sys.out.println("3 Byte val (250)= " + 
          myfm.getValue(adr3, 1));

        int adr4;
        FI2Sys.out.println("4 alloca Byte = " + 
          (adr4=myfm.newLoc(1)));
        FI2Sys.out.println("5 free 2 adr=" + adr2);
        myfm.freeLoc(adr2, 2);
        int adr6;
        FI2Sys.out.println("6 alloca Byte = " + 
          (adr6=myfm.newLoc(1)));
        int adr7;
        FI2Sys.out.println("7 alloca Word = " + 
          (adr7=myfm.newLoc(2)));

        FByte fb1 = new FByte(myfm);
        FI2Sys.out.println("fb1 Byte adr= " + fb1.adr);
        fb1.wr(130);
        FI2Sys.out.println("fb1 (130)= " + fb1.rd());
        FWord fw1 = new FWord(myfm);
        FI2Sys.out.println("fw1 Word adr= " + fw1.adr);
        fw1.wr(-30000);
        FI2Sys.out.println("fw1 (-30000)= " + fw1.rd());
        FDWord fd1 = new FDWord(myfm);
        FI2Sys.out.println("fd1 DWord adr= " + fd1.adr);
        fd1.wr(0x12345678);
        FI2Sys.out.println("fd1 (0x12345678)= " + fd1.rd());


        FI2Sys.out.println("8 free 1 adr=" + adr1);
        myfm.freeLoc(adr1, 4);
        FI2Sys.out.println("9 free 3 adr=" + adr3);
        myfm.freeLoc(adr3, 1);
        FI2Sys.out.println("10 free 4 adr=" + adr4);
        myfm.freeLoc(adr4, 1);
        FI2Sys.out.println("11 free 6 adr=" + adr6);
        myfm.freeLoc(adr6, 1);
        FI2Sys.out.println("12 free 7 adr=" + adr7);
        myfm.freeLoc(adr7, 2);

        fd1.finalize();
        fb1.wr(0x81);
        FI2Sys.out.println("Prima di finalize fb1 (0x81)= " + 
          fb1.rd());
        fb1.finalize();
        FI2Sys.out.println(
  "Dopo finalize, la lettura dovrebbe sollevare eccezione!");
        FI2Sys.out.println("Dopo finalize fb1 = " + 
          fb1.rd());
    } //[m][s] main

} //{c} FileMem
